import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np
import matplotlib.pyplot as plt

torch.cuda.empty_cache()
π = np.pi
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

dataset_storage_path = '/content/drive/My Drive/torch-dataset/'
dataset_storage_path = '/home/doek/torch-dataset'

from anyact import nn_freeze, optim_lr
from chartjs import ChartLine

#############################################################
### TOY PROBLEM 1 ###########################################
#############################################################

def toy_problem(x,proba_flip=0.05):
	out = (x[:,0]>0) ^ (x[:,1]>0)
	flip = (torch.rand(x.shape[0])<proba_flip).to(x.device)
	return (out & flip) | (torch.logical_not(out) & torch.logical_not(flip))
	# return out
	# return (x[:,1]**2-2*x[:,0]>0) ^ (x[:,1]**3-x[:,0]**2>0)

def toy_viz_surf(model):
	device = model.device
	res = 2**8
	x = torch.linspace(-2,2,res)
	y = torch.linspace(-2,2,res)
	grid_x, grid_y = torch.meshgrid((x,y))
	coords = torch.stack([grid_x,grid_y])
	coords = coords.permute(1,2,0).view(-1,2).to(device)
	L = nn_freeze(nn.Linear(2,3).to(device))
	L.weight.data = torch.tensor([[.5,0],[0,0],[0,.5]]).to(device)
	L.bias.data = torch.zeros(3).to(device)
	pixels = L(model(coords)).view(res,res,3)
	plt.imshow(pixels.cpu().detach(),extent=(-2, 2, -2, 2),origin='lower')

def toy_viz(model,toy_set_x,toy_set_y):
	toy_nn_y = torch.argmax(model(toy_set_x),dim=1)
	toy_nn_y_err = (1*toy_set_y - toy_nn_y) == 0
	plt.figure(figsize=(20,6))
	plt.subplot(1,3,1)
	plt.scatter(toy_set_x[ toy_set_y][:,0].cpu(), toy_set_x[ toy_set_y][:,1].cpu(),c='red',s=2)
	plt.scatter(toy_set_x[~toy_set_y][:,0].cpu(), toy_set_x[~toy_set_y][:,1].cpu(),c='blue',s=2)
	toy_viz_surf(lambda x: 1.*F.one_hot(1*toy_problem(x,proba_flip=0)))
	plt.title("ground truth + noised samples")
	plt.subplot(1,3,2)
	plt.scatter(toy_set_x[toy_nn_y==1][:,0].cpu(), toy_set_x[toy_nn_y==1][:,1].cpu(),c='red',s=2)
	plt.scatter(toy_set_x[toy_nn_y==0][:,0].cpu(), toy_set_x[toy_nn_y==0][:,1].cpu(),c='blue',s=2)
	plt.title("model's perception + relabeled samples")
	toy_viz_surf(model)
	plt.subplot(1,3,3)
	# plt.scatter(toy_set_x[:,0].cpu(), toy_set_x[:,1].cpu(),c='white',s=2)
	plt.scatter(toy_set_x[~toy_nn_y_err][:,0].cpu(), toy_set_x[~toy_nn_y_err][:,1].cpu(),c='green',s=2)
	toy_viz_surf(lambda x: 1. - torch.abs(1.*F.one_hot(1*toy_problem(x,proba_flip=0))-model(x)))
	plt.title("difference between model and ground truth + label mismatch")

def toy_train(model,epochs,lr=1e-4,viz=True,toy_set_x=None,toy_set_y=None):
	device = model.device
	if toy_set_x==None: toy_set_x = torch.randn((1000,2)).to(device)
	if toy_set_y==None: toy_set_y = toy_problem(toy_set_x).to(device)
	if not hasattr(model, 'my_logger'): model.my_logger = ChartLine('loss', ['train'])
	if not hasattr(model, 'my_epoch'): model.my_epoch = 0
	if not hasattr(model, 'my_optimizer'): model.my_optimizer = torch.optim.Adam(model.parameters(), lr=lr)
	toy_opti = model.my_optimizer
	optim_lr(toy_opti,lr)
	# plt.figure(figsize=(20,5))
	for epoch in range(epochs): # it's more like sample count as dataset is generated live
		model.my_epoch+=1
		# shape = [batch_size,2]
		# x = torch.randn(shape).to(model.device)
		x = toy_set_x
		model.train()
		toy_opti.zero_grad()
		loss = F.mse_loss(model(x), 1.*F.one_hot(1*toy_problem(x,proba_flip=0)))
		loss.backward()
		toy_opti.step()
		# print(loss.item())
		# plt.scatter(epoch,np.log(loss.item()),s=2)
		if viz: model.my_logger.append(model.my_epoch, {'train': np.log(loss_tot/counter)})
	# infos
	N = 1000
	toy_set_test_x = torch.randn((N,2)).to(device)
	toy_set_test_y = toy_problem(toy_set_test_x).to(device)
	toy_nn_test_y = torch.argmax(model(toy_set_test_x),dim=1)
	correct_count = torch.sum(toy_nn_test_y-toy_set_test_y*1==0)
	if viz: toy_viz(model,toy_set_test_x,toy_set_test_y)
	print(f"correct rate : {correct_count/N}")
	return correct_count/N

# toy_nn = nn.Sequential(*[
#   nn.Linear(2,16),nn.ReLU(),
#   # nn.Linear(16,16),nn.ReLU(),
#   # nn.Linear(16,16),nn.ReLU(),
#   # nn.Linear(16,16),nn.ReLU(),
#   # nn.Linear(16,16),nn.ReLU(),
#   nn.Linear(16,2),nn.Softmax(),
# ]).to(device)
# toy_train(toy_nn,2**12,lr=1e-2)

#############################################################
### MNIST ###################################################
#############################################################

import torchvision
from torchvision import transforms
from torchvision import datasets
from torchvision.transforms import ToTensor
from torchvision.transforms.functional import InterpolationMode

def shared_mnist_train(model,epochs,batch_size,lr,viz,set_train_x,set_train_y,device):
	model.train()
	if not hasattr(model, 'my_optimizer'): model.my_optimizer = torch.optim.Adam(model.parameters(), lr=1e-4)
	model_opti = model.my_optimizer
	optim_lr(model_opti,lr)
	# if viz: plt.figure(figsize=(20,10))
	if not hasattr(model, 'my_logger'): model.my_logger = ChartLine('loss', ['train'])
	if not hasattr(model, 'my_epoch'): model.my_epoch = 0
	# if viz: model.my_logger.show()
	for epoch in range(epochs):
		model.my_epoch+=1
		loss_tot = 0
		counter = 0
		l = len(set_train_x)
		for i in range(0,l,batch_size):
			x = set_train_x[i:min(l-1,i+batch_size)].to(device)
			y = set_train_y[i:min(l-1,i+batch_size)].to(device)
			model_opti.zero_grad()
			loss = F.mse_loss(model(x), y)
			loss.backward()
			model_opti.step()
			loss_tot += loss.item()
			counter += 1
		# if viz: plt.scatter(epoch,np.log(loss_tot/counter),s=2)
		if viz: model.my_logger.append(model.my_epoch, {'train': np.log(loss_tot/counter)})
	# if viz: plt.xscale('log')
	# shared_mnist_perf(model,set_train_x,set_train_y,set_test,set_test_y)

def shared_mnist_perf(model,name,set_x,set_y,batch_size = 1024,device='cpu'):
	# N = len(set_x)
	# y    = torch.argmax(      set_y , dim=1)
	# y_nn = torch.argmax(model(set_x), dim=1)
	# correct_count = torch.sum(y-y_nn==0).item()
	# print(f"correct rate on {name} set = {correct_count/N}")
	###################################
	#### avoid crash filling memory ###
	###################################
	correct_rate = 0
	counter = 0
	N=0
	l = len(set_x)
	for i in range(0,l,batch_size):
		N+=1
		n = min(l-1,i+batch_size)-i
		x = set_x[i:i+n].to(device)
		y = set_y[i:i+n].to(device)
		y_re = torch.argmax(      y , dim=1)
		y_nn = torch.argmax(model(x), dim=1)
		correct_rate += torch.sum(y_re-y_nn==0).item() / n
	print(f"correct rate on {name} set = {correct_rate/N}")

def mnist_train(model,epochs,batch_size=1000,lr=1e-4,viz=True,device='cpu'):
	# device = model.device
	mnist_set_train   = datasets.MNIST(root=dataset_storage_path, download=True, train=True , transform=ToTensor())
	mnist_set_train   = mnist_set_train.data.float().view(-1,1,28,28)/255.
	mnist_set_train_y = F.one_hot(mnist_set_train.targets).float()
	shared_mnist_train(model,epochs,batch_size,lr,viz,mnist_set_train,mnist_set_train_y,device=device)
	shared_mnist_perf(model,"train",mnist_set_train,mnist_set_train_y,device=device)
	mnist_set_test  = datasets.MNIST(root=dataset_storage_path, download=True, train=False, transform=ToTensor())
	mnist_set_test    = mnist_set_test .data.float().view(-1,1,28,28)/255.
	mnist_set_test_y  = F.one_hot(mnist_set_test .targets).float()
	shared_mnist_perf(model,"test" ,mnist_set_test ,mnist_set_test_y,device=device)

def fashion_mnist_train(model,epochs,batch_size=1000,lr=1e-4,viz=True,device='cpu'):
	# device = model.device
	fashion_mnist_set_train = datasets.FashionMNIST(root=dataset_storage_path, download=True, train=True , transform=ToTensor())
	fashion_set_train   = fashion_mnist_set_train.data.float().view(-1,1,28,28)/255.
	fashion_set_train_y = F.one_hot(fashion_mnist_set_train.targets).float()
	shared_mnist_train(model,epochs,batch_size,lr,viz,fashion_set_train,fashion_set_train_y,device=device)
	shared_mnist_perf(model,"train",fashion_set_train,fashion_set_train_y,device=device)
	fashion_mnist_set_test  = datasets.FashionMNIST(root=dataset_storage_path, download=True, train=False, transform=ToTensor())
	fashion_set_test    = fashion_mnist_set_test .data.float().view(-1,1,28,28)/255.
	fashion_set_test_y  = F.one_hot(fashion_mnist_set_test .targets).float()
	shared_mnist_perf(model,"test" ,fashion_set_test ,fashion_set_test_y,device=device)


